/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

/*
 *
 * ATmega128L Emulator
 *
 */


#include "defs.h"

#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "srec.h"
#include "atmega128.h"
#include "iomanager.h"
#include "debug.h"
#include "parse_config.h"
#include "table.h"

static char sreg_flags[9] = "CZNVSHTI";
static int interrupt=0;
static int currchip=0;
static int nchips=0;
static int testing = 0;
static int backend = 0;
int verbosity=V_NORMAL;

static void signal_handler(int type){
    signal(type, SIG_IGN);
    
    switch(type){
        case SIGINT:
            interrupt=1;
            break;
        case SIGCHLD:
            wait(NULL);
            break;
        default:
            printf("signal handler received unknown signal\n");
            break;
    }
    signal(type, signal_handler);
}

static void execute_chips(atmega_chip **chips, uint64_t currentcycle){
    int i;
    for(i=0; i<nchips; i++){
        /* execute current */
        if(chips[i]->cycles == currentcycle) {
            chips[i]->mode=ATEMU_MODE_EXECUTE;
            atmega_step( chips[i] );
        }
        atmega_ioman_tick_clocks(chips[i]);
    }
}

static void decode_chips(atmega_chip **chips, uint64_t currentcycle){
    int i;
    
    for(i=0; i<nchips; i++){
        /* try and show where we will be, come next clock tick */
        if(chips[i]->am_sleeping){
            /* we are sleeping, assume we will continue */
            chips[i]->mode=ATEMU_MODE_DECODE;
            chips[i]->pc--;
            atmega_step( chips[i] );
            chips[i]->pc++;
        }else if(chips[i]->cycles == currentcycle){
            /* we broke right before the gicen instruction */
            chips[i]->mode=ATEMU_MODE_DECODE;
            atmega_step( chips[i] );
        }else{
            /* we are currently executing, and will continue to
             * execute this in the next cycle */
            uint32_t pc;
            uint64_t cycles;

            /* save values */
            pc=chips[i]->pc;
            cycles=chips[i]->cycles;

            /* move into the past */
            chips[i]->pc=chips[i]->oldpc;
            chips[i]->cycles=chips[i]->oldcycles;

            /* decode */
            chips[i]->mode=ATEMU_MODE_DECODE;
            atmega_step( chips[i] );

            /* restore values */
            chips[i]->pc=pc;
            chips[i]->cycles=cycles;
        }
    }
}

static void dump( atmega_chip *chip, uint8_t *mem, uint16_t length, uint16_t begin )
{
    int addr, inchar;
    char memchar;

    mem += begin;

    for( addr = 0; addr < length; addr += 16 )
    {
        /* so I sort of cheat... */
        log_dump(chip, "");

        printf( "%06x:    ", addr + begin );
        printf( "%02x %02x %02x %02x %02x %02x %02x %02x  ",
            mem[addr], mem[addr + 1], mem[addr + 2], mem[addr + 3],
            mem[addr + 4], mem[addr + 5], mem[addr + 6], mem[addr + 7] );
        printf( "%02x %02x %02x %02x %02x %02x %02x %02x",
            mem[addr + 8], mem[addr + 9], mem[addr + 10], mem[addr + 11],
            mem[addr + 12], mem[addr + 13], mem[addr + 14], mem[addr + 15] );
        printf("    ");
        for(inchar = 0; inchar < 16; inchar++) {
            memchar = mem[addr + inchar];
            putchar(((memchar < 32) || (memchar > 126)) ? '.' : memchar);
        }
        printf("\n");
    }

    fflush(stdout);
}

static void unassemble( atmega_chip *chip, char *file ){
    uint32_t oldpc;
    int delta;
    int nnops=0;
    
    oldpc=chip->pc;
    chip->pc=0;
    chip->mode=ATEMU_MODE_DECODE;
    set_decode_mode(DECODE_DUMP, file);

    do{
        /* cheat around the fetch stage, to avoid having to print
         * thousands of NOPs at the end of the program */
        if(chip->flash[chip->pc]==0x0000){
            nnops++;
            delta=1;
            chip->pc++;
        }else{
            if(nnops>0){
                int i;
                chip->pc-=nnops;
                for(i=0; i<nnops; i++){
                    log_opcode(chip, "NOP\n");
                    chip->pc++;
                }
                nnops=0;
            }
            delta=atmega_step(chip);
            if(delta == -1){
                /* illegal instruction, do it as bytes */
                log_opcode(chip, ".byte\t0x%02x, 0x%02x\n", BOTTOM_BYTE(chip->flash[chip->pc]), TOP_BYTE(chip->flash[chip->pc]));
                delta=1;
            }
            chip->pc+=delta;
        }
    }while(chip->pc < chip->flash_words && (delta == 1 || delta == 2) && !interrupt);

    chip->pc=oldpc;
    set_decode_mode(DECODE_NORMAL, NULL);
}

static void dump_regs( atmega_chip *chip )
{
    char sreg[9];
    int i;
    int reg;

    for ( i = 0; i < 8; i++ )
        sreg[i] = ( GET_FLAG( *chip->sreg, 1 << i ) ) ? 
            (sreg_flags[i]) : (' ');
    sreg[8] = 0;

    printf( "[%lld] PC: 0x%08x\tSREG: 0x%02x[%s]\tSTACK: 0x%02x%02x\n",
        chip->cycles, chip->pc * 2, *chip->sreg, sreg, *chip->sph, *chip->spl );

    for( reg = 0; reg < 32; reg++ )
    {
        printf( "r%2d: %02x  ", reg, chip->register_file[reg] );

        if (! ( (reg + 1) % 8 ) )  printf( "\n" );
    }

    printf( "\n\n" );
}



static void dump_flash( atmega_chip *chip, uint32_t dump_addr , uint32_t dump_bytes)
{
    dump( chip, (uint8_t*)chip->flash, dump_bytes, dump_addr );
}

static void dump_sram( atmega_chip *chip, uint32_t dump_addr , uint32_t dump_bytes)
{
    dump( chip, (uint8_t*)chip->sram, dump_bytes, dump_addr );
}

static void has_changed_npoint( void *a, void *extra )
{
    watch_t *w = (watch_t *) a;
    atmega_chip *chip = (atmega_chip *) extra;
    uint8_t curr_val;

    atmega_read_sram(chip, w->addr, &(curr_val));

    if ( curr_val != w->value )
    {
        w->value = curr_val;

        printf( "notify addr %x id %d = %x, chip %d\n", w->addr, w->id,
            w->value, chip->id );
        fflush( stdout );
    }
}

static void mainloop( atmega_chip **chips )
{
    int quit = 0;
    int breakset = 0;
    int watchset = 0;
    uint64_t currentcycle = 0;
    int i;
    char input[1024];

    for(i=0; i<nchips; i++){
        chips[i]->mode=ATEMU_MODE_DECODE;
        atmega_step( chips[i] );
    }

    while( !quit )
    {
        printf( "[atemu]('?' for help) " );
        if(backend){
            /* hackish...... */
            printf("\n");
        }
        fflush(stdout);

        if ( fgets( input, sizeof(input), stdin) ) 
        {
            if ( toupper(input[0]) == 'Q' )
            {
                quit = 1;
            }
            else if ( toupper(input[0]) == '?' ){
                printf("Commands (case-insensitive):\n");
                printf("\tS or Enter \t\tStep one instruction\n");
                printf("\tD address [length]\tDump the SRAM\n");
                printf("\tF address [length]\tFlash dump\n");
                printf("\tU [output file]\t\tUnassemble program\n");
                printf("\tC chip number\t\tChip select\n");
                printf("\tB address\t\tBreakpoint at the Hex Address\n");
                printf("\tW address\t\tWatchpoint at the Hex Address\n");
                printf("\tN address\t\tNotfiypoint at the Hex Address\n"); 
                printf("\tX B|W|N addr\t\tDelete Break, Watch or Notify point at addr\n");
                printf("\tG\t\t\tGo until breakpoint\n");
                printf("\tQ\t\t\tQuit\n");
            }
            else if ( toupper(input[0]) == 'C' )
            {
                sscanf( input, "%*s %d", &currchip );
                if( currchip < 0)
                    currchip = 0;
                if( currchip >= nchips)
                    currchip = nchips - 1;
                printf( "Current chip is now %d.\n", currchip );
            }
            else if ( toupper(input[0]) == 'S' || input[0]=='\n' )
            {
                int i;

                /* run current instruction */
                execute_chips(chips, currentcycle);
                currentcycle++;

                /* check notify points */
                for( i = 0; i < nchips; i++ ) {
                    table_foreach( chips[i]->notifytable,
                        has_changed_npoint, chips[i] );
                }

                for( i = 0; i < nchips; i++ ) {
                    if ( chips[i]->stack_changed != 0 ) {
                        chips[i]->stack_changed = 0;
                        stack_dump( chips[i]->call_stack, chips[i]->id );
                    }
                }

                /* decode next instruction */
                decode_chips(chips, currentcycle);
            }
            else if ( toupper(input[0]) == 'F' )
            {
                uint32_t dumpaddr;
                uint32_t dumpbytes;
                int numargs;
                
                numargs=sscanf( input, "%*s %x %x", &dumpaddr, &dumpbytes );
                dumpaddr=(dumpaddr/16)*16;
                
                if(numargs == 1){
                    dump_flash( chips[currchip], dumpaddr, 16 );
                }else if(numargs == 2){
                    dump_flash( chips[currchip], dumpaddr, dumpbytes );
                }else{
                    printf("Incorrect usage of 'F' command\n");
                }
            }
            else if ( toupper(input[0]) == 'J' )
            {
                uint32_t jmpaddr;
                
                sscanf( input, "%*s %x", &jmpaddr );
                chips[currchip]->pc = jmpaddr / 2;
            }
            else if ( toupper(input[0]) == 'P' )
            {
                uint32_t pokeaddr;
                uint32_t pokebyte;
                
                sscanf( input, "%*s %x %x", &pokeaddr, &pokebyte);
                chips[currchip]->sram[pokeaddr] = pokebyte;
            }
            else if ( toupper(input[0]) == 'R' )
            {
                dump_regs( chips[currchip]);
            }
            else if ( toupper(input[0]) == 'D' )
            {
                uint32_t dumpaddr;
                uint32_t dumpbytes;
                int numargs;
                
                numargs=sscanf( input, "%*s %x %x", &dumpaddr, &dumpbytes );
                dumpaddr=(dumpaddr/16)*16;
                
                if(numargs == 1){
                    dump_sram( chips[currchip], dumpaddr, 16 );
                }else if(numargs == 2){
                    dump_sram( chips[currchip], dumpaddr, dumpbytes );
                }else{
                    printf("Incorrect usage of 'D' command\n");
                }
            }
            else if ( toupper(input[0]) == 'U')
            {
                int count=0;
                char file[1024];
                
                count=sscanf(input, "%*s %1023s", file);

                if(count < 1){
                    unassemble(chips[currchip], NULL);
                }else{
                    unassemble(chips[currchip], file);
                }
            }
            else if ( toupper(input[0]) == 'B' )
            {
                break_t *bpoint;

                bpoint=malloc(sizeof(*bpoint));
                breakset++;
                bpoint->id=breakset;
                sscanf( input, "%*s %x", &(bpoint->addr) );
                log_debug(chips[currchip], V_QUIET, "breakpoint %x set.\n", bpoint->addr);
                bpoint->addr = bpoint->addr / 2;
                table_store(chips[currchip]->breaktable, bpoint);
            }
            else if ( toupper(input[0]) == 'W' )
            {
                watch_t *wpoint;

                wpoint=malloc(sizeof(*wpoint));
                watchset++;
                wpoint->id=watchset;
                sscanf( input, "%*s %x", &(wpoint->addr) );
                log_debug(chips[currchip], V_QUIET, "watchpoint %x set.\n", wpoint->addr );
                atmega_read_sram(chips[currchip], wpoint->addr, &(wpoint->value));
                table_store(chips[currchip]->watchtable, wpoint);
            }
            else if ( toupper(input[0]) == 'N' )
            {
                watch_t *wpoint;

                wpoint=malloc(sizeof(*wpoint));
                sscanf( input, "%*s %x %d", &(wpoint->addr), &(wpoint->id));

                atmega_read_sram(chips[currchip], wpoint->addr, &(wpoint->value));
                table_store(chips[currchip]->notifytable, wpoint);

                printf( "notify addr %x id %d = %x, chip %d\n", wpoint->addr, 
                    wpoint->id, wpoint->value, chips[currchip]->id );
            }
            else if ( toupper(input[0]) == 'X' )
            {
                char type;
                uint32_t addr;

                sscanf( input, "%*s %c %x", &type, &addr);
                if(toupper(type) == 'B'){
                    break_t query;
                    query.addr = addr/2;
                    if(table_delete(chips[currchip]->breaktable, &query, free) >= 0){
                        log_debug(chips[currchip], V_QUIET, "breakpoint cleared.\n");
                    }else{
                        log_debug(chips[currchip], V_QUIET, "no such breakpoint.\n");
                    }
                }else if (toupper(type) == 'W'){
                    watch_t query;
                    query.addr = addr;
                    if(table_delete(chips[currchip]->watchtable, &query, free) >= 0){
                        log_debug(chips[currchip], V_QUIET, "watchpoint cleared.\n");
                    }else{
                        log_debug(chips[currchip], V_QUIET, "no such watchpoint.\n");
                    }
                }else if (toupper(type) == 'N'){
                    watch_t query;
                    query.addr = addr;
                    if(table_delete(chips[currchip]->notifytable, &query, free) >= 0){
                        log_debug(chips[currchip], V_QUIET, "notifypoint cleared.\n");
                    }else{
                        log_debug(chips[currchip], V_QUIET, "no such notifypoint.\n");
                    }
                }
            }
            else if ( toupper(input[0]) == 'G' )
            {
                int breakhit=0; /* have we hit a brealpoint? */

                /* clear interrupt flag, if user pressed ctrl-c */
                interrupt=0;

                do{
                    execute_chips(chips, currentcycle);

                    /* if testing, break on a NOP */
                    if(testing && chips[0]->flash[chips[0]->pc]==0){
                        interrupt=1;
                        printf("TEST END\n");
                    }
                    currentcycle++;

#ifdef SPEEDTEST
                    if(currentcycle >= 100000000){
                        interrupt=1;
                        quit=1;
                    }
#endif

                    /* determine how many breakpoints have been hit */
                    if(breakset > 0 || watchset > 0){
                        break_t *bpoint, query;
                        for(i=0; i < nchips; i++){
                            if(chips[i]->cycles == currentcycle && !chips[i]->am_sleeping){
                                query.addr=chips[i]->pc;
                                bpoint=table_find(chips[i]->breaktable, &query);
                                if(bpoint!=NULL){
                                    breakhit++;
                                    log_debug(chips[i], V_QUIET, "breakpoint %x hit.\n", 2 * bpoint->addr);
                                }
                                if(chips[i]->watchhit != 0){
                                    chips[i]->watchhit=0;
                                    breakhit++;
                                }
                            }
                        }
                    }
                }
                while ( !interrupt && !breakhit );

                if(!testing)
                    decode_chips(chips, currentcycle);

                for(i=0; i < nchips; i++) {
                    table_foreach( chips[i]->notifytable,
                        has_changed_npoint, chips[i] );
                    stack_dump( chips[i]->call_stack, chips[i]->id );
                }
            }
        }
    }
}

int main( int argc, char **argv )
{
    int cmderror = 0;
    error_t result;
    atmega_chip **chips;
    char *fname=NULL;
    char *config="mica2.conf";
    char optstring[] = "htvqbcV";
    char option;
    int i;

    while( (option = getopt(argc, argv, optstring))!=EOF){
        switch (option){
            case 'c':
                /* single-run: machine config */
                if(optarg){
                    config = optarg;
                }
                else{
                    printf("option c requires a file name\n");
                    cmderror=1;
                }
                break;
            case 't':
                testing = 1;
                break;
            case 'b':
                backend = 1;
                break;
            case 'h':
                printf("Options:\n");
                printf("\t-c <filename>\tspecify config file.  Default is mica2.conf\n");
                printf("\t-t\t\trun in 'test mode'\n");
                printf("\t-b\t\trun in 'backend mode' (output altered for xatdb)\n");
                printf("\t-v\t\tincreases verbosity by one each time.\n");
                printf("\t-q\t\tdecreases verbosity by one each time.\n");
                printf("\t-V\t\tdisplay version information\n");
                printf("\t-h\t\tDisplay this list.\n\n");

                exit(0);
                break;
            case 'v':
                verbosity++;
                if(verbosity > V_ALL){
                    verbosity=V_ALL;
                }
                break;
            case 'q':
                verbosity--;
                if(verbosity < V_QUIET){
                    verbosity=V_QUIET;
                }
                break;
            case 'V':
                printf( "[ATmega128L Emulator %s]\n", PACKAGE_VERSION );
                exit(0);
                break;
            case ':':
                printf("missing required parameter for an option\n");
                cmderror=1;
                break;
            case '?':
                printf("unknown option\n");
                cmderror=1;
                break;
            default:
                printf("Unknown getopt return code '%c'\n", option);
                cmderror=1;
                break;
        }
    }

    if(optind < argc){
        fname = argv[optind];
    }
    
    if(cmderror || (fname == NULL)){
        printf("Usage: %s [-options] filename\n", argv[0]);
        printf("\tUse -h for option list\n");
        exit(1);
    }

    printf( "[ATmega128L Emulator %s]\n", PACKAGE_VERSION );

    /* set up the chips */
    nchips = atemu_load_file( &chips, fname, config );

    if ( nchips <= 0 ) {
        printf("Usage: %s [-options] filename.srec\n", argv[0]);
        printf("\tUse -h for option list\n");
        exit(-1);
    }

/*
    if(nchips==1 && config!=NULL && srec!=NULL){
        chips=malloc(sizeof(*chips));
        result=atmega_load_chip(chips, 0, config, srec);
    }else if(runfile!=NULL){
        result=atmega_parse_runfile(&chips, &nchips, runfile);
    }else{
*/

    signal(SIGINT, signal_handler);
    signal(SIGCHLD, signal_handler);
    set_decode_mode(DECODE_NORMAL, NULL);

    mainloop( chips );

    for(i=0; i<nchips; i++){
        result = atmega_destroy_chip( &chips[i] );
        if ( result < 0 )
        {
            printf( "[error] atmega_destroy_%d chip failed [%d]\n", i, result );
            return result;
        }
        printf( "Destroyed chip %d.\n", i ); 
    }

    printf( "Done.\n" ); 

    return 0;
}
